/* Copyright (c) 2012 The Chromium Authors. All rights reserved.
 * Use of this source code is governed by a BSD-style license that can be
 * found in the LICENSE file.
 */

#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <mono/jit/jit.h>
#include <mono/metadata/assembly.h>
#include <mono/metadata/debug-helpers.h>

#include "mono_helpers.h"
#include "my_assert.h"
#include "post_message.h"
#include "url_loader.h"

extern void mono_set_corlib_data(void *data, size_t size);

void post_message_icall(MonoString* arg) {
  post_message(mono_string_to_utf8(arg));
}

/* Actually initialize Mono, the data being passed in is mscorlib.dll downloaded
 * to a buffer which will not be deallocated (required by mono_set_corlib_data).
 */
void finish_mono_initialization(int success, const char* buffer, size_t size,
                                const void* data) {
  assert(success);
  /* Native Client specific, this is so init doesn't try the filesystem. */
  mono_set_corlib_data((char*)buffer, size);
  /* libmono is relocated from where it was originally built, so the default
   * directories for lib and etc won't work.  We're going to treat all files we
   * load for mono as being at the root of the filesystem, "/".
   */
  mono_set_dirs("/", "/");
  mono_jit_init((const char*)data);
  /* This requires an implementation of 'post_message_icall' in the .cs file. */
  mono_add_internal_call(CS_CLASS_NAME"::post_message_icall",
                         post_message_icall);

  /* Signals to JS that we can load assemblies. */
  post_message("Mono Initialized");
}

/* This function doesn't actually initialize Mono, it kicks off the download of
 * mscorlib.dll which is needed for initialization.  On download,
 * finish_mono_initialization will be called.
 */
void initialize_mono(const char* domain) {
  char *domain_copy = strdup(domain);
  load_file_from_url("mscorlib.dll", finish_mono_initialization, domain_copy);
}

/* Given a buffer containing an assembly, load it from memory in Mono.
 * Note: The buffer was allocated by url_loader code and will not be free'd.
 */
void load_assembly_from_memory(int success, const char* buffer, size_t size,
                               const void* data) {
  char assembly_name[256];
  snprintf(assembly_name, 256, "/%s", (const char*)data);

  MonoImageOpenStatus imgstatus = MONO_IMAGE_ERROR_ERRNO;
  MonoImage* mi =
      mono_image_open_from_data_with_name(
          (char*)buffer,
          size,
          0, /* need_copy: set to 0 because the buffer will never free */
          &imgstatus,
          0, /* refonly */
          assembly_name);
  assert(mi && imgstatus == MONO_IMAGE_OK);
  imgstatus = MONO_IMAGE_OK; /* not written on success */
  MonoAssembly* ma =
      mono_assembly_load_from_full(
          mi,
          assembly_name,
          &imgstatus,
          0 /* refonly */);
  assert(ma && imgstatus == MONO_IMAGE_OK);

  /* Signals to JS that an assembly has been loaded. */
  char buf[256];
  snprintf(buf, 256, "Loaded Assembly %s", assembly_name);
  post_message(buf);
  free((void*)data);
}

/* Kick off the download of an assembly, to be handled by
 * load_assembly_from_memory above.
 */
void load_assembly(const char* url, const char* assembly_name) {
  load_file_from_url(url, load_assembly_from_memory, strdup(assembly_name));
}

/* Execute the static Main method from the loaded assembly. */
void execute_assembly(const char* assembly_name) {
  MonoAssemblyName* man = mono_assembly_name_new(assembly_name);
  assert(man);
  MonoAssembly* ma = mono_assembly_loaded(man);
  assert(ma);
  char *args[1];
  args[0] = (char*)assembly_name;
  
  mono_jit_exec(mono_domain_get(), ma, 1 /* argc */, args);
}

/* Execute the named method, a static method which takes one string argument. */
void run_static_mono_method_string_arg(const char* assembly_name,
                                       const char* method,
                                       const char* arg) {
  MonoAssemblyName* man = mono_assembly_name_new(assembly_name);
  assert(man);
  MonoAssembly* ma = mono_assembly_loaded(man);
  assert(ma);
  MonoImage* mi = mono_assembly_get_image(ma);
  assert(mi);
  MonoMethodDesc* mmd = mono_method_desc_new(method, 0 /* include_namespace */);
  assert(mmd);
  MonoMethod* mm = mono_method_desc_search_in_image(mmd, mi);
  assert(mm);
  void *args[1];
  args[0] = mono_string_new(mono_domain_get(), arg);

  mono_runtime_invoke(mm, NULL /* "this" object */, args, NULL /* exception */);
}

