/* Copyright (c) 2012 The Chromium Authors. All rights reserved.
 * Use of this source code is governed by a BSD-style license that can be
 * found in the LICENSE file.
 */

#include <stdio.h>
#include <string.h>

#include "ppapi/c/pp_errors.h"
#include "ppapi/c/pp_module.h"
#include "ppapi/c/pp_var.h"
#include "ppapi/c/ppb.h"
#include "ppapi/c/ppb_core.h"
#include "ppapi/c/ppb_instance.h"
#include "ppapi/c/ppb_messaging.h"
#include "ppapi/c/ppb_url_loader.h"
#include "ppapi/c/ppb_url_request_info.h"
#include "ppapi/c/ppb_var.h"
#include "ppapi/c/ppp.h"
#include "ppapi/c/ppp_instance.h"
#include "ppapi/c/ppp_messaging.h"

#include "../common/mono_helpers.h"
#include "../common/post_message.h"
#include "../common/url_loader.h"

static PPB_GetInterface g_get_browser = NULL;
static PPB_Var* g_ppb_var_interface = NULL;

/* Note: result is owned by caller. */
char* VarToCStr(struct PP_Var var) {
  uint32_t len = 0;
  if (g_ppb_var_interface != NULL) {
    const char* var_c_str = g_ppb_var_interface->VarToUtf8(var, &len);
    if (len > 0) {
      char* c_str = (char*)malloc(len + 1);
      memcpy(c_str, var_c_str, len);
      c_str[len] = '\0';
      return c_str;
    }
  }
  return NULL;
}

/**
 * Called when the NaCl module is instantiated on the web page. This function
 * initializes some helper modules for loading data from URLs and reporting
 * progress and diagnostic messages back to Javascript.
 */
static PP_Bool Instance_DidCreate(PP_Instance instance,
                                  uint32_t argc,
                                  const char* argn[],
                                  const char* argv[]) {
  initialize_url_loader(instance,
                        g_get_browser(PPB_URLLOADER_INTERFACE),
                        g_get_browser(PPB_URLREQUESTINFO_INTERFACE),
                        g_get_browser(PPB_CORE_INTERFACE),
                        g_get_browser(PPB_VAR_INTERFACE));
  initialize_post_message(instance,
                          g_get_browser(PPB_MESSAGING_INTERFACE),
                          g_get_browser(PPB_CORE_INTERFACE),
                          g_get_browser(PPB_VAR_INTERFACE));
  initialize_mono(PROJECT_NAME".exe");
  return PP_TRUE;
}

static void Instance_DidDestroy(PP_Instance instance) {
}

static void Instance_DidChangeView(PP_Instance instance,
                                   PP_Resource view_resource) {
}

static void Instance_DidChangeFocus(PP_Instance instance,
                                    PP_Bool has_focus) {
}

static PP_Bool Instance_HandleDocumentLoad(PP_Instance instance,
                                           PP_Resource url_loader) {
  return PP_FALSE;
}

/**
 * Handler that gets called as a result of Javascript PostMessage.
 *
 * This function will execute a command, either loading a Mono assembly or
 * executing managed code inside of Mono.
 */

static void Messaging_HandleMessage(PP_Instance instance,
                                    struct PP_Var message) {
  if (message.type != PP_VARTYPE_STRING)
    return;

  const char* kLoadAssemblyId = "loadAssembly";
  const char* kExecuteAssemblyId = "executeAssembly";
  const char* kRunMethodId = "runMethod";

  char* msg = VarToCStr(message);
  if (strncmp(msg, kLoadAssemblyId, strlen(kLoadAssemblyId)) == 0) {
    /* Load a mono assembly from a given URL */
    /* Example: loadAssembly hw.exe */
    char *url = msg + strlen(kLoadAssemblyId) + 1;
    char *assembly_name = url;
    char *last_slash = strrchr(kLoadAssemblyId, '/');
    if (last_slash) {
      assembly_name = last_slash + 1;
    }
    load_assembly(url, assembly_name);
  }
  else if (strncmp(msg, kExecuteAssemblyId, strlen(kExecuteAssemblyId)) == 0) {
    /* Execute a loaded assembly. */
    /* Example: executeAssembly hw */
    char *assembly_name = msg + strlen(kExecuteAssemblyId) + 1;
    execute_assembly(assembly_name);
  }
  else if (strncmp(msg, kRunMethodId, strlen(kRunMethodId)) == 0) {
    /* Run a particular method from an assembly. */
    /* Example: runMethod hw HelloWorld:ReverseText(string) Foo! */

    /* Very simple parsing to extract the method name and argument.
     * Modify the string in place to chop up the string.
     */
    char *assembly_name = msg + strlen(kRunMethodId) + 1;
    char *method_name = strchr(assembly_name, ' ');
    method_name[0] = '\0';
    method_name += 1;
    char *arg = strchr(method_name, ' ');
    arg[0] = '\0';
    arg += 1;

    run_static_mono_method_string_arg(assembly_name, method_name, arg);
  }

  free(msg);
}

/**
 * Entry points for the module.
 */
PP_EXPORT int32_t PPP_InitializeModule(PP_Module a_module_id,
                                       PPB_GetInterface get_browser) {
  g_ppb_var_interface = (PPB_Var*)(get_browser(PPB_VAR_INTERFACE));
  g_get_browser = get_browser;
  return PP_OK;
}

PP_EXPORT const void* PPP_GetInterface(const char* interface_name) {
  if (strcmp(interface_name, PPP_INSTANCE_INTERFACE) == 0) {
    static PPP_Instance instance_interface = {
      &Instance_DidCreate,
      &Instance_DidDestroy,
      &Instance_DidChangeView,
      &Instance_DidChangeFocus,
      &Instance_HandleDocumentLoad,
    };
    return &instance_interface;
  }
  else if (strcmp(interface_name, PPP_MESSAGING_INTERFACE) == 0) {
    static PPP_Messaging messaging_interface = {
      &Messaging_HandleMessage,
    };
    return &messaging_interface;
  }
  return NULL;
}

PP_EXPORT void PPP_ShutdownModule() {
}

