/* Copyright (c) 2012 The Chromium Authors. All rights reserved.
 * Use of this source code is governed by a BSD-style license that can be
 * found in the LICENSE file.
 */

#include <string.h>

#include "ppapi/c/pp_completion_callback.h"
#include "ppapi/c/pp_errors.h"
#include "ppapi/c/pp_var.h"
#include "ppapi/c/ppb_core.h"
#include "ppapi/c/ppb_url_loader.h"
#include "ppapi/c/ppb_url_request_info.h"
#include "ppapi/c/ppb_var.h"

#include "my_assert.h"
#include "post_message.h"
#include "url_loader.h"


static PP_Instance g_instance;
static const PPB_URLLoader* g_url_loader;
static const PPB_URLRequestInfo* g_url_request;
static const PPB_Core* g_core;
static const PPB_Var* g_var;

void initialize_url_loader(PP_Instance instance, const void *url_loader,
                           const void *url_request,const void *core,
                           const void *var) {
  assert(url_loader && url_request && core && var);
  g_instance = instance;
  g_url_loader = (const PPB_URLLoader*)url_loader;
  g_url_request = (const PPB_URLRequestInfo*)url_request;
  g_core = (const PPB_Core*)core;
  g_var = (const PPB_Var*)var;
}

struct loader_data {
  char *buf;
  size_t size;
  size_t alloced;
  url_callback user_cb;
  const void* user_data;
  PP_Resource loader;
};

const size_t BUFFER_READ_SIZE = 4096;

void read_some(void* data);

static void read_callback(void* data, int32_t result) {
  struct loader_data *ldata = (struct loader_data*)data;
  if (result == PP_OK) {
    /* We're done reading the file. */
    ldata->user_cb(1, ldata->buf, ldata->size, ldata->user_data);
    free(ldata);
  } else if (result > 0) {
    /* 'result' bytes were read into memory. */
    ldata->size += (size_t)result;
    read_some(data);
  } else {
    ldata->user_cb(0, NULL, 0, ldata->user_data);
  }
}

/* Read up to BUFFER_READ_SIZE bytes more from the URLLoader.
 * Allocate more space in the destination buffer if needed.
 */
void read_some(void* data) {
  struct loader_data *ldata = (struct loader_data*)data;
  while (ldata->alloced < ldata->size + BUFFER_READ_SIZE) {
    /* The buffer isn't big enough to hold BUFFER_READ_SIZE more bytes. */
    char *temp = (char*)malloc(ldata->alloced * 2);
    memcpy(temp, ldata->buf, ldata->size);
    free(ldata->buf);
    ldata->buf = temp;
    ldata->alloced *= 2;
  }

  struct PP_CompletionCallback cb =
      PP_MakeCompletionCallback(read_callback, data);
  int32_t read_ret = g_url_loader->ReadResponseBody(
      ldata->loader,
      ldata->buf + ldata->size,
      BUFFER_READ_SIZE,
      cb);
  assert(read_ret == PP_OK_COMPLETIONPENDING);
}

static void open_callback(void* data, int32_t result) {
  assert(result == PP_OK);
  read_some(data);
}

void load_file_from_url(const char* url, url_callback user_cb,
                        const void* user_data) {
  PP_Resource request = g_url_request->Create(g_instance);
  struct PP_Var url_var = g_var->VarFromUtf8(url, strlen(url));
  g_url_request->SetProperty(request, PP_URLREQUESTPROPERTY_URL, url_var);
  g_var->Release(url_var);
  PP_Resource loader = g_url_loader->Create(g_instance);

  struct loader_data* data =
      (struct loader_data*)malloc(sizeof(struct loader_data));
  data->buf = (char*)malloc(BUFFER_READ_SIZE);
  data->size = 0;
  data->alloced = BUFFER_READ_SIZE;
  data->user_cb = user_cb;
  data->user_data = user_data;
  data->loader = loader;

  struct PP_CompletionCallback cb =
      PP_MakeCompletionCallback(open_callback, data);
  int32_t open_ret = g_url_loader->Open(loader, request, cb);
  assert(open_ret == PP_OK_COMPLETIONPENDING);
  g_core->ReleaseResource(request);
}

